<?php
/**
 * @file
 * Contains \Drupal\supercookie\Controller\SupercookieController.
 */

namespace Drupal\supercookie\Controller;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpFoundation\Session\Session;
use Symfony\Component\HttpFoundation\JsonResponse;
use Drupal\Core\Path\AliasManager;
use Drupal\Component\Utility\Unicode;
use Drupal\supercookie\SupercookieManager;

/**
 * Controller routines for Supercookie resources.
 */
class SupercookieController implements ContainerInjectionInterface {

  /**
   * TODO.
   */
  protected $supercookieManager, $config, $requestStack, $session, $aliasManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('supercookie.manager'),
      $container->get('request_stack'),
      $container->get('session'),
      $container->get('path.alias_manager')
    );
  }
  /**
   * {@inheritdoc}
   */
  public function __construct(SupercookieManager $supercookie_manager, RequestStack $request_stack, Session $session, AliasManager $alias_manager) {
    $this->supercookieManager = $supercookie_manager;
    $this->config = $this->supercookieManager->config;
    $this->requestStack = $request_stack;
    $this->session = $session;
    $this->aliasManager = $alias_manager;
  }

  /**
   * TODO.
   */
  private function instance() {

    $path_json = $this->aliasManager->getAliasByPath('/supercookie');
    $request = $this->requestStack->getCurrentRequest();
    $ip_addr = $request->getClientIp();
    $args = $request->query->all();
    $supercookie = NULL;

    if ($request->getPathInfo() == $path_json) {
      if (!empty($args['client']) && !empty($args['date'])) {

        // Set client + server data.
        $data = array(
          'server' => array(
            'REMOTE_ADDR' => $ip_addr,
            'REMOTE_HOST' => gethostbyaddr($ip_addr),
            'HTTP_USER_AGENT' => $request->server->get('HTTP_USER_AGENT'),
            'HTTP_ACCEPT' => $request->server->get('HTTP_ACCEPT'),
          ),
          'client' => $args['client'],
        );

        $hash = md5(serialize($data));

        // Upsert supercookie instance.
        $supercookie = $this->supercookieManager
          ->match($hash)
          ->save($args['date']);

        $this->session->set($this->config['supercookie_name_server'], $hash);

        // TODO: Track entity views; also note the "ref" arg's controlling JS var.
        // @see Drupal.behaviors.supercookie.entitiesTracked
        if (!empty($args['ref'])) {
          /*
          // Get the current node menu object.
          $menu_item = menu_get_item(drupal_get_normal_path($args['ref']));

          // Gather nid and tid values from current node entity.
          if (!empty($menu_item['map'][0]) && $menu_item['map'][0] == 'node') {
            // Get a list of all taxonomy_term fields.
            $fields_term = &drupal_static(__FUNCTION__ . '_fields_term');
            if (empty($fields_term)) {
              $fields = field_info_field_map();
              $fields_term = array();
              foreach ($fields as $field_name => &$meta) {
                if ($meta['type'] == 'taxonomy_term_reference') {
                  $fields_term[] = $field_name;
                }
              }
            }

            $node = $menu_item['map'][1];
            $tid = array();
            foreach ($fields_term as $field_name) {
              if (!empty($node->{$field_name})) {
                $items = field_get_items('node', $node, $field_name);
                foreach ($items as &$term) {
                  $tid[] = $term['tid'];
                }
              }
            }
            // Increment entity view counts.
            $this->supercookie = $this->supercookie->trackNodes(array($node->nid));
            if (!empty($tid)) {
              $this->supercookie = $this->supercookie->trackTerms($tid);
            }
          }
          */
        }
      }
    }

    if (empty($supercookie)) {
      $hash = $this->session->get($this->config['supercookie_name_server']);
      $supercookie = $this->supercookieManager
        ->match($hash);
    }

    // Now set header, cookie and JSON for client. Note explicit header() here
    // vs. attached to JsonResponse below...we want to apply this header to
    // all responses, not just the /supercookie response.
    $once = &drupal_static(get_class($this) . Unicode::ucwords(__FUNCTION__), FALSE);
    if (empty($once)) {
      if (empty($supercookie->dnt)) {
        $request->server->set('HTTP_' . str_replace('-', '_', Unicode::strtoupper($this->config['supercookie_name_header'])), $supercookie->data);
        header($supercookie->config['supercookie_name_header'] . ': ' . $supercookie->data, TRUE);
        setcookie($supercookie->config['supercookie_name_server'], $supercookie->data, $supercookie->expires, '/');
      }
      else {
        header($supercookie->config['supercookie_name_header'] . ': ' . '', TRUE);
        setcookie($supercookie->config['supercookie_name_server'], '', 0);
      }

      $once = TRUE;
    }

    $data = array(
      'dnt' => !empty($supercookie->dnt),
      'scid' => $supercookie->scid,
      'hash' => $supercookie->data,
      'expires' => $supercookie->expires,
      'json' => $path_json,
      'name_server' => $supercookie->config['supercookie_name_server'],
      'name_header' => $supercookie->config['supercookie_name_header'],
    );

    $headers = array();
    if ($request->getPathInfo() == $path_json) {
      $headers = array(
        'Cache-Control' => 'no-cache, no-store, must-revalidate',
        'Pragma' => 'no-cache',
        'Expires' => '0',
      );
    }

    return new JsonResponse($data, 200, $headers);
  }
  /**
   * Menu callback to dump a JSON blob of all current, transformed supercookie data.
   */
  public function render() {
    return $this->instance();
  }

  /**
   * Gets a human-readable array from raw supercookie values.
   */
  private function report_format($object) {

    $mongo = $this->config['supercookie_mongodb'];
    $account = user_load($object->uid);

    $human = array(
      'user' => (!empty($account->name) ? $account->name : t('anonymous')),
      'cookie' => $object->scid,
      'hash' => $object->data,
      'created' => format_date($object->created, 'e'),
      'modified' => format_date(REQUEST_TIME, 'e'),
      'expires' => format_date($object->expires, 'e'),
    );

    // Try to unserialize row blobs.
    if (!$mongo) {
      $tid = unserialize($object->tid);
      $nid = unserialize($object->nid);
      $custom = unserialize($object->custom);

      if ($tid !== FALSE) {
        $object->tid = $tid;
      }
      if ($nid !== FALSE) {
        $object->nid = $nid;
      }
      if ($custom !== FALSE) {
        $object->custom = unserialize($object->custom);
        $human['custom'] = $object->custom;
      }
    }

    // Add term names + counts to response.
    if (!empty($object->tid)) {
      $human['terms'] = taxonomy_term_load_multiple(array_keys($object->tid));
      foreach ($human['terms'] as &$term) {
        $term = (object) array(
          $term->name => $object->tid[$term->tid],
        );
      }
      $human['terms'] = array_values($human['terms']);
    }

    // Add node titles + counts to response.
    if (!empty($object->nid)) {
      $human['nodes'] = node_load_multiple(array_keys($object->nid));
      foreach ($human['nodes'] as &$node) {
        $node = (object) array(
          $node->title => $object->nid[$node->nid],
        );
      }
      $human['nodes'] = array_values($human['nodes']);
    }

    return $human;
  }
  /**
   * Menu callback to dump a JSON blob of all current, transformed supercookie data.
   */
  public function report() {

    $data = array();
    $results = \Drupal::database()
      ->select('supercookie', 'sc')
      ->fields('sc', array(
        'scid',
        'uid',
        'created',
        'modified',
        'expires',
        'data',
        'tid',
        'nid',
        'custom',
      ))
      ->execute();

    while ($row = $results->fetch()) {
      $human = $this->report_format($row);

      // Allow other modules to customize each row as needed (e.g. per its own
      // usage of the "custom" field).
      \Drupal::moduleHandler()->alter('supercookie.admin_report', $human);

      $data[] = $human;
    }

    return new JsonResponse($data, 200, array(
      'Cache-Control' => 'no-cache, no-store, must-revalidate',
      'Pragma' => 'no-cache',
      'Expires' => '0',
    ));
  }

}
